#!/bin/sh
# This script must be able to run on busybox-powered devices, so use POSIX shell instead of bash.

### BEGIN INIT INFO
# Provides:          networkoptix-mediaserver
# Required-Start:    networking
# Required-Stop:
# Default-Start:     2 3 4 5
# Default-Stop:      0 1 6
# Short-Description: Network Optix Nx Go Media Server
### END INIT INFO

#--------------------------------------------------------------------------------------------------
# Default values for the configuration variables

CUSTOMIZATION="networkoptix"

SERVICE_NAME="mediaserver"

# Settings for killing the Server if it uses too much memory.
SECONDS_TO_WAIT_BEFORE_KILL_9=120
KILL_MEDIASERVER_IF_HIGH_MEM_USAGE=1
HIGH_MEM_USAGE_KEEP_FREE_KB=44272 #< Total memory minus the maximum allowed Server memory.

MEDIASERVER_DIR="/"
LIBS_DIR="${MEDIASERVER_DIR}/lib"
BIN_NAME="mediaserver"
BIN_FILE="${MEDIASERVER_DIR}/bin/${BIN_NAME}"
LOGS_DIR="${MEDIASERVER_DIR}/var/log"

# ATTENTION: The args should not contain spaces, because busybox shell does not support arrays.
RUN_ARGS="-e"

CORE_PATH="${MEDIASERVER_DIR}/cores"

# Total memory in the system.
MEM_TOTAL_KB="$(cat /proc/meminfo | grep 'MemTotal' | awk '{print $2}')"

MAX_ALLOWED_MEM_USAGE_KB="$(expr "${MEM_TOTAL_KB}" - "${HIGH_MEM_USAGE_KEEP_FREE_KB}")"

# Touch .flag file to capture all Server output to a file.
MEDIASERVER_OUT_FLAG="${LOGS_DIR}/mediaserver-out.flag"
MEDIASERVER_OUT_LOG="${LOGS_DIR}/mediaserver-out.log"

# Touch .flag file to capture all output of this script to a file.
OUT_FLAG="${LOGS_DIR}/$(basename "$0")-out.flag"
OUT_LOG="${LOGS_DIR}/$(basename "$0")-out.log"

# Prefix to the command that runs the Server; if not empty, should end with a space.
START_COMMAND_PREFIX=""

scriptDir="$(dirname "$0")"

#--------------------------------------------------------------------------------------------------
# Include the configuration file for the flavor-specific variables

. "${scriptDir}/mediaserver.sh.conf"

#--------------------------------------------------------------------------------------------------
# Other global variables

. "${scriptDir}/../build_info.txt"
FULL_VERSION="${fullVersion}"

#--------------------------------------------------------------------------------------------------

# Redirect all output of this script to the log file.
redirect_output() # log_file
{
    log_file="$1"

    # If the old log file exists, rename it to first available "filename_#.log", stopping at 9.
    old_log_file_message=""
    if [ -f "${log_file}" ]; then
        for i in $(seq 1 9)
        do
            old_log_file="${log_file%.*}_$i.log" #< Remove the extension and add the new suffix.
            if [ ! -f "${old_log_file}" ]; then #< Unused filename found.
                break
            fi
        done
        mv "${log_file}" "${old_log_file}"
        old_log_file_message=" (old log file renamed to $(basename "${old_log_file}"))"
    fi

    echo "$0: See the log in ${log_file}${old_log_file_message}"

    exec 1<&- #< Close stdout fd.
    exec 2<&- #< Close stderr fd.
    exec 1<>"${log_file}" #< Open stdout as $LOG_FILE for reading and writing.
    exec 2>&1 #< Redirect stderr to stdout.
}

log_args() # "$@"
{
    echo "Script [$0] started with pid $$ (parent pid ${PPID}), with $# arg(s):"
    echo "{"
    for arg in "$@"; do
        echo "    [${arg}]"
    done
    echo "}"
    echo ""
}

process_core_dump_if_needed()
{
    mkdir -p "${CORE_PATH}"

    core_orig="${CORE_PATH}/core"
    time=$(date +"%s")

    # Do not process the core if the previous core has appeared less than a day before.
    old_core=$(ls "${core_orig}".* 2>/dev/null)
    old_core_time="${old_core#$core_orig.}" #< Remove "core." prefix.
    # Check whether exactly one core file exists with a time in a name.
    if [ "$(expr match "${old_core_time}" "[0-9]*\$")" != 0 ]; then
        if [ "$(expr ${time} - ${old_core_time})" -gt "$(expr 24 \* 60 \* 60)" ]; then
            echo "Old core is more than a day old"
            return 1
        fi
    fi

    core="${core_orig}.${time}"
    if ! mv "${core_orig}" "${core}" 2>/dev/null; then
        echo "No core file found at ${core_orig}"
        return 1
    fi

    ls "${core_orig}".* | grep -v "${core}" | xargs rm 2>/dev/null #< Delete all other core files.

    return 0
}

stop_server() # [signal]
{
    signal="$1"

    server_pid="$(pidof "${BIN_NAME}")"
    if [ -z "${server_pid}" ]; then
        echo "${SERVICE_NAME} is not running"
        return 1
    fi

    echo "Stopping Server (${BIN_NAME})..."
    kill ${signal} "${server_pid}" >/dev/null 2>&1

    i=0
    while [ ! -z "$(pidof "${BIN_NAME}")" ] && [ "$i" -lt "${SECONDS_TO_WAIT_BEFORE_KILL_9}" ]; do
        sleep 1
        i=$((i + 1))
    done

    if [ ! -z "$(pidof "${BIN_NAME}")" ]; then
        echo "Server did not stop after $i seconds; doing killall -9"
        kill -9 "${server_pid}" >/dev/null 2>&1
    fi

    echo "Done"
}

prepare_server_to_run()
{
    echo "Preparing server to run..."

    if [ "${USER}" = "root" ]; then
        # If Mediaserver runs under unpreviliged users, the management should be done in the
        # installer.
        fallback_libs_script="${MEDIASERVER_DIR}/lib/fallback/manage_fallback_libs.sh"
        if [ -f "${fallback_libs_script}" ]; then
            "${fallback_libs_script}"
        fi
    fi
}

run_server_in_background() # output_file
{
    output_file="$1";

    start_command="${START_COMMAND_PREFIX}\"${BIN_FILE}\" ${RUN_ARGS} >>\"${output_file}\" 2>&1"

    echo "Starting Server with command: ${start_command}"
    /bin/sh -c "${start_command}; echo \"Server process has finished with exit status [\$?]\"" &
    sleep 3

    # See if the Server instantly crashes.
    if [ -z "$(pidof "${BIN_NAME}")" ]; then
        echo "FAILED"
        return 1
    fi
    echo "${SERVICE_NAME} has been started"
}

start_server()
{
    server_pid=$(pidof "${BIN_NAME}")
    if [ ! -z "${server_pid}" ]; then
       echo "${SERVICE_NAME} is already running with pid ${server_pid}"
       return 1
    fi

    process_core_dump_if_needed

    if [ -f "${MEDIASERVER_OUT_FLAG}" ]; then
        mkdir -p "$(dirname "${MEDIASERVER_OUT_LOG}")"
        output_file="${MEDIASERVER_OUT_LOG}"
    else
        output_file="/dev/null"
    fi

    prepare_server_to_run
    run_server_in_background "${output_file}"
}

get_pid_by_cmd_line_fragment()
{
    cmd_line="$1"
    pgrep -f "${cmd_line}" 2>/dev/null
    if [ "$?" = "127" ]; then #< "pgrep" is not found, trying to get PID using ps.
        # "COLUMNS=2048" is needed to avoid stripping ps output.
        COLUMNS=2048 ps | grep -v grep | grep "${cmd_line}" | sed -e 's/^\s*//' | cut -d" " -f1
    fi
}

print_watchdog_pid()
{
    get_pid_by_cmd_line_fragment "mediaserver.sh run_watchdog"
}

stop_server_if_high_mem_usage()
{
    server_pid=$(pidof "${BIN_NAME}")
    if [ -z "${server_pid}" ]; then
        exit 0
    fi

    mem_usage_kb="$(cat "/proc/${server_pid}/status" \
        | grep VmRSS | cut -d ":" -f2 | tr -d "\t " | cut -d "k" -f 1)"

    if [ "${mem_usage_kb}" -gt "${MAX_ALLOWED_MEM_USAGE_KB}" ]; then
        echo "High mem usage detected: ${mem_usage_kb} KB; stopping Server..."
        stop_server
    fi
}

run_watchdog()
{
    while true; do
        if [ -z "$(pidof "${BIN_NAME}")" ]; then
            echo "${SERVICE_NAME} watchdog: Restarting ${BIN_NAME}..."
            start_server
        fi

        stop_server_if_high_mem_usage

        sleep 20
    done
}

start_watchdog()
{
    watchdog_pid="$(print_watchdog_pid)"
    if [ ! -z "${watchdog_pid}" ]; then
        echo "${SERVICE_NAME} watchdog is already running with pid ${watchdog_pid}"
        return
    fi

    "$0" "run_watchdog" >/dev/null 2>&1 &
    echo "${SERVICE_NAME} watchdog has been started"
}

stop_watchdog()
{
    watchdog_pid="$(print_watchdog_pid)"
    if [ -z "${watchdog_pid}" ]; then
        echo "${SERVICE_NAME} watchdog is not running"
        return 1
    fi

    echo "Stopping ${SERVICE_NAME} watchdog..."
    kill "${watchdog_pid}"
    sleep 1
    watchdog_pid=$(print_watchdog_pid)
    if [ ! -z "${watchdog_pid}" ]; then
        echo "${SERVICE_NAME} watchdog still not killed; doing kill -9"
        kill -9 "${watchdog_pid}"
    fi
    echo "Done"
}

setup_system()
{
    ulimit -n $(cat /proc/sys/fs/file-max 2>/dev/null || echo 8192)
    ulimit -c unlimited
}

print_status()
{
    watchdog_pid=$(print_watchdog_pid)
    if [ ! -z "${watchdog_pid}" ]; then
        echo "${SERVICE_NAME} watchdog is running with pid ${watchdog_pid}"
    else
        echo "${SERVICE_NAME} watchdog is stopped"
    fi

    service_pid=$(pidof "${BIN_NAME}")
    if [ ! -z "${service_pid}" ]; then
        echo "${SERVICE_NAME} is running with pid ${service_pid}"
        exit 0
    else
        echo "${SERVICE_NAME} is stopped"
        exit 1
    fi
}

help()
{
    cat \
<<EOF
Usage: $0 <command>
Supported <command> values:
 start # Start ${SERVICE_NAME} and watchdog, outputting to ${MEDIASERVER_OUT_LOG} if
 ${MEDIASERVER_OUT_FLAG} exists.
 stop [<signal>] # Stop watchdog and kill ${SERVICE_NAME} with SIGTERM or <signal>.
 status # Report whether ${SERVICE_NAME} and watchdog are running, and their pids. Exit status
     is non-zero if stopped.
 restart, reload, force-reload # Stop the watchdog and ${SERVICE_NAME}, start ${SERVICE_NAME} and
     the watchdog.
 run_watchdog # Run the watchdog in the foreground.
 stop_watchdog # Stop the watchdog, if it is running.
EOF
}

#--------------------------------------------------------------------------------------------------

main()
{
    if [ -f "${OUT_FLAG}" ]; then
        redirect_output "${OUT_LOG}"
        log_args "$@"
        set -x #< Log each command.
    fi

    command=$1
    if [ $# -gt 1 ]; then
        shift
    fi

    case "${command}" in
        start)
            setup_system
            start_server && start_watchdog
            ;;
        stop)
            stop_watchdog
            stop_server $1
            ;;
        run_watchdog)
            setup_system
            run_watchdog
            ;;
        stop_watchdog)
            stop_watchdog
            ;;
        status)
            print_status
            ;;
        restart|reload|force-reload)
            stop_watchdog
            stop_server
            setup_system
            start_server && start_watchdog
            ;;
        *)
            echo "ERROR: Unsupported command \"{$command}\"." >&2
            help
            exit 1
    esac
}

main "$@"
